'Copyright (c) 2011 Donald Milne
'All rights reserved.
'
'This file is part of AOSkills3.
'
'AOSkills3 is free software: you can redistribute it and/or modify
'it under the terms of the GNU General Public License as published by
'the Free Software Foundation, either version 3 of the License, or
'(at your option) any later version.
'
'AOSkills3 is distributed in the hope that it will be useful,
'but WITHOUT ANY WARRANTY; without even the implied warranty of
'MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'GNU General Public License for more details.
'
'You should have received a copy of the GNU General Public License
'along with AOSkills3. If not, see <http://www.gnu.org/licenses/>.

Imports System
Imports System.IO
Imports System.Reflection

Public Class ActionLogger

' Creates a per character log file in the logs sub-folder of the application folder
' Allows changes to a character to be logged

    Private prvLogFile As StreamWriter

    Private prvLogStatus As String

    Public Sub New(ByRef strCharName As String, ByVal intLevel As Integer)
        Dim strAppDir As String = Path.GetDirectoryName(Assembly.GetExecutingAssembly().Location)
        prvLogFile = New StreamWriter(strAppDir & Path.DirectorySeparatorChar & "logs" & Path.DirectorySeparatorChar & strCharName & ".log", True)
        prvLogFile.WriteLine("")
        prvLogFile.WriteLine("+--------------------------------------------------------------------")
        prvLogFile.WriteLine("| Character : " & strCharName & " at " & CStr(intLevel))
        prvLogFile.WriteLine("| Logging Started: " & System.DateTime.Now.ToString("dd-MMM-yyyy HH:mm:ss"))
        prvLogFile.WriteLine("+--------------------------------------------------------------------")
        prvLogStatus = "ENABLED"
    End Sub

    Public Sub Close()
        If Not IsNothing(prvLogFile) Then
            ' set focus to objFrmMain
            prvLogFile.Close()
            prvLogFile = Nothing
        End If
    End Sub

    Public Sub EnableLogging()
        prvLogStatus = "ENABLED"
    End Sub

    Public Sub DisableLogging()
        prvLogStatus = "DISABLED"
    End Sub

    Private Sub Log(ByRef strText As String)
        If prvLogStatus = "ENABLED" Then
            prvLogFile.WriteLine(strText)
        End If
    End Sub

    Public Sub LogIP(ByRef strSkill As String, ByVal intOldIP As Integer, intNewIP As Integer)
        Dim strText As String = "CHANGED SKILL " & strSkill & " IP from " & CStr(intOldIP) & " to " & CStr(intNewIP)
        Me.Log(strText)
    End Sub

    Public Sub LogPerk(ByRef strPerkName As String, ByVal intOldChk As Boolean, ByVal intNewChk As Boolean)
        Dim strText As String = "CHANGED PERK " & strPerkName & " from " & CStr(intOldChk) & " to " & CStr(intNewChk)
        Me.Log(strText)
    End Sub

    ' use this also for Symbiants, Spirits and Alien Eye
    Public Sub LogEquipment(Byref strSlot As String, ByRef strOldEquipment As String, ByVal intOldQL As Integer, ByRef strNewEquipment As String, ByVal intNewQL As Integer)
        Dim strText As String = "CHANGED ITEM IN " & strSlot & " from " & isSelStr(strOldEquipment) & " @QL " & CStr(intOldQL) & " to " & isSelStr(strNewEquipment) & " @QL " & CStr(intNewQL)
        Me.Log(strText)
    End Sub

    Public Sub LogImplant(ByRef strSlot As String, ByRef strOldShinyCluster As String, ByRef strOldBrightCluster As String, ByRef strOldFadedCluster As String, ByVal intOldQL As Integer, ByRef strNewShinyCluster As String, ByRef strNewBrightCluster As String, ByRef strNewFadedCluster As String, ByVal intNewQL As Integer)
        Dim strText As String = "CHANGED IMPLANT IN " & strSlot & " from " & isSelStr(strOldShinyCluster) & "/" & isSelStr(strOldBrightCluster) & "/" & isSelStr(strOldFadedCluster) & " @QL " & CStr(intOldQL) & " to " & isSelStr(strNewShinyCluster) & "/" & isSelStr(strNewBrightCluster) & "/" & isSelStr(strNewFadedCluster) & " @QL " & CStr(intNewQL)
        Me.Log(strText)
    End Sub

    Public Sub LogLevel(ByVal intOldLvl As Integer, ByVal intNewLvl As Integer)
        Dim strText As String = "Changed Level from " & CStr(intOldLvl) & " to " & CStr(intNewLvl)
        Me.Log(strText)
    End Sub

    Private Function isSelStr(ByRef strItem As String) As String
        If strItem = SEL_STRING Then
            Return "Empty"
        Else
            Return strItem
        End If
    End Function

End Class
