'Copyright (c) 2003-2012 Donald Milne
'All rights reserved.
'
'This file is part of AOSkills3.
'
'AOSkills3 is free software: you can redistribute it and/or modify
'it under the terms of the GNU General Public License as published by
'the Free Software Foundation, either version 3 of the License, or
'(at your option) any later version.
'
'AOSkills3 is distributed in the hope that it will be useful,
'but WITHOUT ANY WARRANTY; without even the implied warranty of
'MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'GNU General Public License for more details.
'
'You should have received a copy of the GNU General Public License
'along with AOSkills3. If not, see <http://www.gnu.org/licenses/>.

Imports System
Imports System.Xml
Imports Microsoft.VisualBasic

Module modXML

    Private professionMap() As Integer = { _
        -1, 8, -1, -1, -1, 0, 13, -1, -1, 3, 10, -1, -1, -1 _
    }

    Public loadedBuffs As New System.Collections.Specialized.StringCollection

    Private Class WrongProfException : Inherits Exception
    End Class

    Private Class WrongBreedException : Inherits Exception
    End Class

    Private Class LevelTooLowException : Inherits Exception
    End Class

    Private Class AILevelTooLowException : Inherits Exception
    End Class

    Public Class xmlBuffError : Inherits Exception
    End Class

    Public Class BadSkillException : Inherits Exception
    End Class

    Public Sub saveDataNew(ByVal filename As String)
        Dim i As Integer
        Dim xwriter As XmlTextWriter
        Dim arrWorking() As String = Nothing
        Dim strSplitString As String = vbCrLf
        Dim caSplitString() As Char = strSplitString.ToCharArray()

        'Create a new XmlTextWriter.
        xwriter = New XmlTextWriter(filename, System.Text.Encoding.UTF8)
        xwriter.Indentation = 4
        xwriter.Formatting = Formatting.Indented
        'Write the beginning of the document including the document declaration. Standalone is true.
        xwriter.WriteStartDocument(True)
        'Write the initial elements
        xwriter.WriteStartElement("CHAR")
        xwriter.WriteElementString("LEVEL", CStr(fmMain.NudLevelValue))
        xwriter.WriteElementString("ALIENTITLE", CStr(fmMain.NudAlienTitleValue))
        xwriter.WriteElementString("PROF", CStr(fmMain.ProfIndex))
        xwriter.WriteElementString("BREED", CStr(fmMain.BreedIndex))
        ' Write Abilities, then Skills
        For i = 0 To NUM_ATTRS - 1
            xwriter.WriteStartElement("ATTR")
            xwriter.WriteAttributeString("ID", CStr(i))
            xwriter.WriteAttributeString("NAME", attrNames(i))
            xwriter.WriteElementString("IP", CStr(attrCA.Item(i).Improvement))
            xwriter.WriteEndElement()
        Next
        For i = 0 To NUM_SKILLS - 1
            xwriter.WriteStartElement("SKILL")
            xwriter.WriteAttributeString("ID", CStr(i))
            xwriter.WriteAttributeString("NAME", skillNames(i))
            xwriter.WriteElementString("IP", CStr(skillsCA.Item(i).Improvement))
            xwriter.WriteEndElement()
        Next
        xwriter.WriteStartElement("PERKS")
        For i = 0 To NUM_PERKS
            ' update to write out in form <PERK NAME="Acrobat" NUM="1" />
            If PerkArray(i).Selected = True Then
                'xwriter.WriteElementString("PERKID", CStr(i))
                xwriter.WriteStartElement("PERK")
                xwriter.WriteAttributeString("NAME", PerkArray(i).PerkName)
                xwriter.WriteAttributeString("NUM", CStr(PerkArray(i).PerkNo))
                xwriter.WriteEndElement()
            End If
        Next
        xwriter.WriteEndElement()
        xwriter.WriteStartElement("AIPERKS")
        For i = 0 To NUM_ALIEN_PERKS
            ' update to write out in form <AIPERK NAME="Shadow" NUM="1" />
            If AlienPerkArray(i).Selected = True Then
                'xwriter.WriteElementString("AIPERKID", CStr(i))
                xwriter.WriteStartElement("AIPERK")
                xwriter.WriteAttributeString("NAME", AlienPerkArray(i).PerkName)
                xwriter.WriteAttributeString("NUM", CStr(AlienPerkArray(i).PerkNo))
                xwriter.WriteEndElement()
            End If
        Next
        xwriter.WriteEndElement()
        xwriter.WriteStartElement("RESEARCH")
        For i = 0 To NUM_RESEARCH
            ' update to write out in form <LEPERK NAME="Assassin's Awareness" NUM="1" />
            If ResearchArray(i).Selected = True Then
                'xwriter.WriteElementString("RESEARCHID", CStr(i))
                xwriter.WriteStartElement("LEPERK")
                xwriter.WriteAttributeString("NAME", ResearchArray(i).PerkName)
                xwriter.WriteAttributeString("NUM", CStr(ResearchArray(i).PerkNo))
                xwriter.WriteEndElement()
            End If
        Next
        xwriter.WriteEndElement()
        xwriter.WriteStartElement("BUFFS")
        For i = 0 To modXML.loadedBuffs.Count - 1
            xwriter.WriteElementString("BUFFNAME", modXML.loadedBuffs(i))
        Next
        xwriter.WriteEndElement()
        xwriter.WriteStartElement("BENEFITS")
        For i = 0 To num_Benefits2
            If BenefitArray2(i).Selected Then
                xwriter.WriteElementString("BENEFITNAME", BenefitArray2(i).BenefitName & " @ QL" & CStr(BenefitArray2(i).SelectedQL))
            End If
        Next
        xwriter.WriteEndElement()
        ' write equipment info
        xwriter.WriteStartElement("EQUIPMENT")
        For i = FIRST_WEAPON_SLOT To FIRST_IMP_SLOT - 1
            xwriter.WriteStartElement("SLOT")
            xwriter.WriteAttributeString("ID", CStr(i))
            xwriter.WriteAttributeString("LO_AOID", CStr(equipment_slots(i, LOW_AOID)))
            xwriter.WriteAttributeString("HI_AOID", CStr(equipment_slots(i, HIGH_AOID)))
            xwriter.WriteAttributeString("QL", CStr(equipment_slots(i, ITEM_QL)))
            xwriter.WriteEndElement()
        Next
        xwriter.WriteEndElement()
        ' write implant info
        xwriter.WriteStartElement("IMPLANTS")
        For i = 0 To MAX_IMPS
            xwriter.WriteStartElement("IMPLANT")
            xwriter.WriteAttributeString("ID", CStr(i))
            xwriter.WriteAttributeString("QL", CStr(impA(i).QL))
            xwriter.WriteAttributeString("TYPE", impA(i).SorI)
            If impA(i).SorI = "I" Then
                xwriter.WriteAttributeString("SHINY", IIf(impA(i).ShinyCluster = SEL_STRING, "Empty", impA(i).ShinyCluster).ToString())
                xwriter.WriteAttributeString("BRIGHT", IIf(impA(i).BrightCluster = SEL_STRING, "Empty", impA(i).BrightCluster).ToString())
                xwriter.WriteAttributeString("FADED", IIf(impA(i).FadedCluster = SEL_STRING, "Empty", impA(i).FadedCluster).ToString())
            ElseIf impA(i).SorI = "S" Then
                xwriter.WriteAttributeString("SYMB", impA(i).SymbiantOrSpirit)
            ElseIf impA(i).SorI = "A" Then
                ' nothing further to do here?
            End If
            xwriter.WriteEndElement()
        Next
        xwriter.WriteEndElement()
        'End the document
        xwriter.WriteEndElement()
        xwriter.WriteEndDocument()
        'Flush the XML document to the underlying stream and close the underlying stream
        xwriter.Close()
    End Sub

    ' load characters from file
    Public Sub loadDataNew(ByVal file_name As String)
        Dim rdr As Xml.XmlTextReader
        Dim node_name As String = ""
        Dim ver As Integer = 1

        Dim buffs As New System.Collections.ArrayList
        Dim perks As New System.Collections.ArrayList
        Dim aiperks As New System.Collections.ArrayList
        Dim research As New System.Collections.ArrayList
        Dim benefits As New System.Collections.ArrayList

        Dim ssAOSInit As aosInitProgress

        Dim level As Integer = -1
        Dim atitle As Integer = -1
        Dim prof As Integer = -1
        Dim breed As Integer = -1
        Dim id As Integer = -1

        Dim ip As Integer = -1

        Dim ab(NUM_ATTRS) As Integer
        Dim sk(NUM_SKILLS) As Integer

        Dim local_equipment_slots(MAX_SLOT, 2) As Integer

        Dim local_implants(13, 6) As String
        Const IMP_QL As Integer = 0
        Const IMP_TYPE As Integer = 1
        Const IMP_SYMB As Integer = 3
        Const IMP_SHINY As Integer = 4
        Const IMP_BRIGHT As Integer = 5
        Const IMP_FADED As Integer = 6

        Dim local_perks(40, 2) As String
        Dim local_aiperks(30, 2) As String
        Dim local_leperks(70, 2) As String
        Const PERK_NAME As Integer = 0
        Const PERK_NUM As Integer = 1
        Dim perks_count As Integer = 0
        Dim aiperks_count As Integer = 0
        Dim leperks_count As Integer = 0

        Dim noError As Boolean = True

        Dim i As Integer

        Try
            loadingData = True
            rdr = New Xml.XmlTextReader(file_name)
            Do While rdr.Read()
                Select Case rdr.NodeType
                    Case Xml.XmlNodeType.Element
                        Select Case rdr.Name
                            Case "ATTR"
                                id = CInt(rdr.GetAttribute("ID"))
                                ip = 0
                            Case "SKILL"
                                id = CInt(rdr.GetAttribute("ID"))
                                ip = 0
                            Case "SLOT"
                                i = CInt(rdr.GetAttribute("ID"))
                                local_equipment_slots(i, LOW_AOID) = CInt(rdr.GetAttribute("LO_AOID"))
                                local_equipment_slots(i, HIGH_AOID) = CInt(rdr.GetAttribute("HI_AOID"))
                                local_equipment_slots(i, ITEM_QL) = CInt(rdr.GetAttribute("QL"))
                            Case "IMPLANT"
                                i = CInt(rdr.GetAttribute("ID"))
                                local_implants(i, IMP_QL) = rdr.GetAttribute("QL")
                                local_implants(i, IMP_TYPE) = rdr.GetAttribute("TYPE")
                                If local_implants(i, IMP_TYPE) = "I" Then
                                    local_implants(i, IMP_SHINY) = rdr.GetAttribute("SHINY")
                                    local_implants(i, IMP_BRIGHT) = rdr.GetAttribute("BRIGHT")
                                    local_implants(i, IMP_FADED) = rdr.GetAttribute("FADED")
                                ElseIf local_implants(i, IMP_TYPE) = "S" Then
                                    local_implants(i, IMP_SYMB) = rdr.GetAttribute("SYMB")
                                End If
                            Case "PERK"
                                local_perks(perks_count, PERK_NAME) = rdr.GetAttribute("NAME")
                                local_perks(perks_count, PERK_NUM) = rdr.GetAttribute("NUM")
                                perks_count = perks_count + 1
                            Case "AIPERK"
                                local_aiperks(aiperks_count, PERK_NAME) = rdr.GetAttribute("NAME")
                                local_aiperks(aiperks_count, PERK_NUM) = rdr.GetAttribute("NUM")
                                aiperks_count = aiperks_count + 1
                            Case "LEPERK"
                                local_leperks(leperks_count, PERK_NAME) = rdr.GetAttribute("NAME")
                                local_leperks(leperks_count, PERK_NUM) = rdr.GetAttribute("NUM")
                                leperks_count = leperks_count + 1
                            Case Else
                                node_name = rdr.Name
                        End Select
                    Case Xml.XmlNodeType.Text
                        Select Case node_name
                            Case "LEVEL"
                                level = CInt(rdr.Value)
                            Case "ALIENTITLE"
                                atitle = CInt(rdr.Value)
                            Case "PROF"
                                prof = CInt(rdr.Value)
                            Case "BREED"
                                breed = CInt(rdr.Value)
                            Case "IP"
                                ip = CInt(rdr.Value)
                            Case "BUFFNAME"
                                buffs.Add(rdr.Value)
                            Case "BENEFITNAME"
                                benefits.Add(rdr.Value)
                        End Select
                    Case Xml.XmlNodeType.EndElement
                        Select Case rdr.Name
                            Case "ATTR"
                                ab(id) = ip
                            Case "SKILL"
                                sk(id) = ip
                        End Select
                End Select
            Loop
        Catch
            noError = False
            MsgBox("There was a problem loading the file <" & file_name & ">", CType(MsgBoxStyle.Exclamation + MsgBoxStyle.OkOnly, Microsoft.VisualBasic.MsgBoxStyle), "File Open Error")
        Finally
            If Not (rdr Is Nothing) Then
                rdr.Close()
            End If
            loadingData = False
        End Try
        If noError = True Then
            loadingData = True
            ResetChar()
            noRecalc = True
            ' put values into system
            If level <> -1 Then
                fmMain.NudLevelValue = level
            End If
            If atitle <> -1 Then
                fmMain.NudAlienTitleValue = atitle
            End If
            If prof <> -1 Then
                fmMain.ProfIndex = prof
            End If
            If breed <> -1 Then
                fmMain.BreedIndex = breed
            End If
            ' force update of limits...
            noRecalc = False
            recalcIP()
            noRecalc = True
            ' abilities
            For i = 0 To NUM_ATTRS - 1
                attrCA.Item(i).Improvement = ab(i)
            Next
            ' process buffs
            For i = 0 To buffs.Count - 1
                loadData_Buff("buffs\" & CStr(buffs(i)) & ".buff")
            Next
            ' process perks
            For i = 0 To perks_count - 1
                ' get ID from name/number and AddPerk that
                AddPerk(GetPerkIndex(local_perks(i, PERK_NAME), CInt(local_perks(i, PERK_NUM))))
            Next
            For i = 0 To aiperks_count - 1
                ' get ID from name/number and AddAlienPerk that
                AddAlienPerk(GetAlienPerkIndex(local_aiperks(i, PERK_NAME), CInt(local_aiperks(i, PERK_NUM))))
            Next
            For i = 0 To leperks_count - 1
                ' get ID from name/number and AddResearch that
                AddResearch(GetResearchIndex(local_leperks(i, PERK_NAME), CInt(local_leperks(i, PERK_NUM))))
            Next
            ' process benefits
            For i = 0 To benefits.Count - 1
                processBenefit(CStr(benefits(i)))
            Next
            ' update perk data
            ConfigurePerks()
            updateCheckboxes()
            updateAICheckboxes()
            updateLECheckboxes()
            ' update org benefit data
            processAllBenefits()
            ' force update of limits...
            noRecalc = False
            recalcIP()
            noRecalc = True
            ' skills
            For i = 0 To NUM_SKILLS - 1
                skillsCA.Item(i).Improvement = sk(i)
            Next
            ' TODO: we need an "is loading" flag so limits can be ignored...
            ' redo equipment modifiers from the array
            For i = FIRST_WEAPON_SLOT To FIRST_IMP_SLOT - 1
                equipment_slots(i, LOW_AOID) = local_equipment_slots(i, LOW_AOID)
                equipment_slots(i, HIGH_AOID) = local_equipment_slots(i, HIGH_AOID)
                equipment_slots(i, ITEM_QL) = local_equipment_slots(i, ITEM_QL)
            Next
            ' setup implants etc
            ssAOSInit = New aosInitProgress
            ssAOSInit.AOSInitMax = impA.Count
            ssAOSInit.lblText.Text = "Update Symbiant List"
            ssAOSInit.Show()
            ssAOSInit.Refresh()
            For i = 0 To MAX_IMPS
                ssAOSInit.AOSInitStep()
                ssAOSInit.Refresh()
                impA(i).UseHandler = False
                impA(i).MaxQL = 300
                impA(i).setType(local_implants(i, IMP_TYPE))
                If impA(i).SorI = "I" Then
                    impA(i).setElement("Shiny", local_implants(i, IMP_SHINY))
                    impA(i).setElement("Bright", local_implants(i, IMP_BRIGHT))
                    impA(i).setElement("Faded", local_implants(i, IMP_FADED))
                    impA(i).MaxQL = CInt(local_implants(i, IMP_QL))
                    impA(i).QL = CInt(local_implants(i, IMP_QL))
                ElseIf impA(i).SorI = "A" Then
                    impA(i).setElement("EOE", "")
                    impA(i).MaxQL = CInt(local_implants(i, IMP_QL))
                    impA(i).QL = CInt(local_implants(i, IMP_QL))
                ElseIf impA(i).SorI = "C" Then
                    impA(i).setElement("CI", "")
                    impA(i).MaxQL = CInt(local_implants(i, IMP_QL))
                    impA(i).QL = CInt(local_implants(i, IMP_QL))
                ElseIf impA(i).SorI = "S" Then
                    impA(i).setSymbsAndSpirits(True)
                    impA(i).setElement("Symbiant", local_implants(i, IMP_SYMB))
                End If
                impA(i).UseHandler = True
            Next
            ssAOSInit.Hide()
            calculateTotalModifiers(combinedSkillMods, combinedAttributeMods, combinedHPNPMods)
            ' set "Other" and "Implant" parts of all skills, abilities etc
            For i = 0 To NUM_SKILLS - 1
                skillsCA.Item(i).Other = combinedSkillMods(i, EQUIP_MODS)
                skillsCA.Item(i).Implant = combinedSkillMods(i, IMP_MODS)
            Next
            For i = 0 To NUM_ATTRS - 1
                attrCA.Item(i).Other = combinedAttributeMods(i, EQUIP_MODS)
                attrCA.Item(i).Implant = combinedAttributeMods(i, IMP_MODS)
            Next
            hpCA.Other = combinedHPNPMods(0, EQUIP_MODS)
            npCA.Other = combinedHPNPMods(1, EQUIP_MODS)
            hpCA.Implant = combinedHPNPMods(0, IMP_MODS)
            npCA.Implant = combinedHPNPMods(1, IMP_MODS)
            fmPDoll.loadIconsFromArray()
            ' final update
            noRecalc = False
            recalcIP()
            loadingData = False
        End If
    End Sub

    ' load characters from profile data
    Public Sub loadRawData(ByVal file_name As String)
        Dim rdr As Xml.XmlTextReader
        Dim node_name As String = ""

        Dim level As Integer = -1
        Dim ailevel As Integer = -1
        Dim prof As Integer = -1
        Dim breed As Integer = -1
        Dim id As Integer = -1

        Dim ip As Integer = -1
        Dim ab(NUM_ATTRS) As Integer
        Dim sk(NUM_SKILLS) As Integer

        Dim noError As Boolean = True

        Dim i As Integer

        Try
            loadingData = True
            rdr = New Xml.XmlTextReader(file_name)
            Do While rdr.Read()
                Select Case rdr.NodeType
                    Case Xml.XmlNodeType.Element
                        Select Case rdr.Name
                            Case "ATTR"
                                id = CInt(rdr.GetAttribute("ID"))
                                ip = 0
                            Case "SKILL"
                                id = CInt(rdr.GetAttribute("ID"))
                                ip = 0
                            Case Else
                                node_name = rdr.Name
                        End Select
                    Case Xml.XmlNodeType.Text
                        Select Case node_name
                            Case "LEVEL"
                                level = CInt(rdr.Value)
                            Case "AILEVEL"
                                ailevel = CInt(rdr.Value)
                            Case "PROF"
                                prof = CInt(rdr.Value)
                            Case "BREED"
                                breed = CInt(rdr.Value)
                            Case "IP"
                                ip = CInt(rdr.Value)
                        End Select
                    Case Xml.XmlNodeType.EndElement
                        Select Case rdr.Name
                            Case "ATTR"
                                ab(id) = ip
                            Case "SKILL"
                                sk(id) = ip
                        End Select
                End Select
            Loop
        Catch
            noError = False
            MsgBox("There was a problem loading the file <" & file_name & ">", CType(MsgBoxStyle.Exclamation + MsgBoxStyle.OKOnly, Microsoft.VisualBasic.MsgBoxStyle), "File Open Error")
        Finally
            If Not (rdr Is Nothing) Then
                rdr.Close()
            End If
            loadingData = False
        End Try
        If noError = True Then
            loadingData = True
            ResetChar()
            noRecalc = True
            ' put values into system
            If level <> -1 Then
                fmMain.NudLevelValue = level
            End If
            If ailevel <> -1 Then
                fmMain.NudAlienTitleValue = ailevel
            End If
            If prof <> -1 Then
                fmMain.ProfIndex = prof
            End If
            If breed <> -1 Then
                fmMain.BreedIndex = breed
            End If
            ' force update of limits...
            noRecalc = False
            recalcIP()
            noRecalc = True
            ' abilities
            For i = 0 To NUM_ATTRS - 1
                attrCA.Item(i).Improvement = ab(i) - breedInit(breed, i)
            Next
            ' force update of limits...
            noRecalc = False
            recalcIP()
            noRecalc = True
            ' skills
            For i = 0 To NUM_SKILLS - 1
                skillsCA.Item(i).Improvement = sk(i) - BASE_SKILL
            Next
            ' final update
            noRecalc = False
            recalcIP()
            loadingData = False
        End If
    End Sub

    ' save character buffs to file
    Public Sub saveData_Buff(ByVal filename As String)
        Dim i As Integer
        Dim xwriter As XmlTextWriter

        'Create a new XmlTextWriter.
        xwriter = New XmlTextWriter(filename, System.Text.Encoding.UTF8)
        xwriter.Indentation = 4
        xwriter.Formatting = Formatting.Indented
        'Write the beginning of the document including the document declaration. Standalone is true.
        xwriter.WriteStartDocument(True)
        'Write the initial elements
        xwriter.WriteStartElement("BUFF")
        ' Write Abilities, then Skills
        For i = 0 To NUM_ATTRS - 1
            If attrCA.Item(i).Buff <> 0 Then
                xwriter.WriteStartElement("ATTR")
                xwriter.WriteAttributeString("ID", CStr(i))
                xwriter.WriteAttributeString("NAME", attrNames(i))
                xwriter.WriteElementString("BUFF", CStr(attrCA.Item(i).Buff))
                xwriter.WriteEndElement()
            End If
        Next
        For i = 0 To NUM_SKILLS - 1
            If skillsCA.Item(i).Buff <> 0 Then
                xwriter.WriteStartElement("SKILL")
                xwriter.WriteAttributeString("ID", CStr(i))
                xwriter.WriteAttributeString("NAME", skillNames(i))
                xwriter.WriteElementString("BUFF", CStr(skillsCA.Item(i).Buff))
                xwriter.WriteEndElement()
            End If
        Next
        If hpCA.Buff <> 0 Then
            xwriter.WriteStartElement("HP")
            xwriter.WriteElementString("BUFF", CStr(hpCA.Buff))
            xwriter.WriteEndElement()
        End If
        If npCA.Buff <> 0 Then
            xwriter.WriteStartElement("NANO")
            xwriter.WriteElementString("BUFF", CStr(npCA.Buff))
            xwriter.WriteEndElement()
        End If
        'End the document
        xwriter.WriteEndElement()
        xwriter.WriteEndDocument()
        'Flush the XML document to the underlying stream and close the underlying stream
        xwriter.Close()
    End Sub

    ' save character perks to file
    Public Sub saveData_Perk(ByVal filename As String)
        Dim i As Integer
        Dim xwriter As XmlTextWriter

        'Create a new XmlTextWriter.
        xwriter = New XmlTextWriter(filename, System.Text.Encoding.UTF8)
        xwriter.Indentation = 4
        xwriter.Formatting = Formatting.Indented
        'Write the beginning of the document including the document declaration. Standalone is true.
        xwriter.WriteStartDocument(True)
        'Write the initial elements
        xwriter.WriteStartElement("PERKLIST")
        xwriter.WriteElementString("PROF", CStr(fmMain.ProfIndex))
        ' should maybe only write this if breed specific perks selected...
        xwriter.WriteElementString("BREED", CStr(fmMain.BreedIndex))
        xwriter.WriteElementString("LEVEL", CStr(fmMain.NudLevelValue))
        xwriter.WriteElementString("AILEVEL", CStr(fmMain.NudAlienTitleValue))
        ' write selected perks
        For i = 0 To NUM_PERKS
            If PerkArray(i).Selected = True Then
                xwriter.WriteStartElement("PERK")
                xwriter.WriteAttributeString("NAME", "", PerkArray(i).PerkName)
                xwriter.WriteString(CStr(PerkArray(i).PerkNo))
                xwriter.WriteEndElement()
            End If
        Next
        For i = 0 To NUM_ALIEN_PERKS
            If AlienPerkArray(i).Selected = True Then
                xwriter.WriteStartElement("AIPERK")
                xwriter.WriteAttributeString("NAME", "", AlienPerkArray(i).PerkName)
                xwriter.WriteString(CStr(AlienPerkArray(i).PerkNo))
                xwriter.WriteEndElement()
            End If
        Next
        For i = 0 To NUM_RESEARCH
            If ResearchArray(i).Selected = True Then
                xwriter.WriteStartElement("RESEARCH")
                xwriter.WriteAttributeString("NAME", "", ResearchArray(i).PerkName)
                xwriter.WriteString(CStr(ResearchArray(i).PerkNo))
                xwriter.WriteEndElement()
            End If
        Next
        'End the document
        xwriter.WriteEndDocument()
        'Flush the XML document to the underlying stream and close the underlying stream
        xwriter.Close()
    End Sub

    ' save character misc to file
    Public Sub saveData_Misc(ByVal filename As String)
        Dim i As Integer
        Dim xwriter As XmlTextWriter

        'Create a new XmlTextWriter.
        xwriter = New XmlTextWriter(filename, System.Text.Encoding.UTF8)
        xwriter.Indentation = 4
        xwriter.Formatting = Formatting.Indented
        'Write the beginning of the document including the document declaration. Standalone is true.
        xwriter.WriteStartDocument(True)
        'Write the initial elements
        xwriter.WriteStartElement("ORGBENEFIT")
        For i = 0 To num_Benefits2
            If BenefitArray2(i).Selected Then
                xwriter.WriteStartElement("ITEM")
                xwriter.WriteAttributeString("QL", CStr(BenefitArray2(i).SelectedQL))
                xwriter.WriteAttributeString("NAME", BenefitArray2(i).BenefitName)
                xwriter.WriteAttributeString("ID", CStr(i))
                xwriter.WriteEndElement()
            End If
        Next
        'End the document
        xwriter.WriteEndElement()
        xwriter.WriteEndDocument()
        'Flush the XML document to the underlying stream and close the underlying stream
        xwriter.Close()
    End Sub

    ' load characters buffs from file
    Public Sub loadData_Buff(ByVal filename As String)
        Dim i As Integer
        Dim ab(NUM_ATTRS) As Integer
        Dim sk(NUM_SKILLS) As Integer
        Dim hp As Integer = 0
        Dim np As Integer = 0
        Dim xreader As XmlTextReader
        Dim noError As Boolean = True
        Dim buff_name As String
        Dim old_noRecalc As Boolean

        For i = 0 To NUM_ATTRS - 1
            ab(i) = 0
        Next
        For i = 0 To NUM_SKILLS - 1
            sk(i) = 0
        Next

        Try
            buff_name = getNamePart(filename)
            'Create a new XmlTextReader.
            xreader = New XmlTextReader(filename)
            xreader.ReadStartElement("BUFF")
            Do While xreader.IsStartElement()
                If xreader.IsStartElement("ATTR") Then
                    i = CInt(xreader.GetAttribute("ID"))
                    xreader.ReadStartElement("ATTR")
                    ab(i) = CInt(xreader.ReadElementString("BUFF"))
                    xreader.ReadEndElement()
                End If
                If xreader.IsStartElement("SKILL") Then
                    i = CInt(xreader.GetAttribute("ID"))
                    xreader.ReadStartElement("SKILL")
                    sk(i) = CInt(xreader.ReadElementString("BUFF"))
                    xreader.ReadEndElement()
                End If
                If xreader.IsStartElement("HP") Then
                    xreader.ReadStartElement("HP")
                    hp = CInt(xreader.ReadElementString("BUFF"))
                    xreader.ReadEndElement()
                End If
                If xreader.IsStartElement("NANO") Then
                    xreader.ReadStartElement("NANO")
                    np = CInt(xreader.ReadElementString("BUFF"))
                    xreader.ReadEndElement()
                End If
            Loop
            xreader.ReadEndElement()
        Catch
            ' do error thing here
            noError = False
            MsgBox("There was a problem loading the buff file <" & filename & ">", CType(MsgBoxStyle.Exclamation + MsgBoxStyle.OKOnly, Microsoft.VisualBasic.MsgBoxStyle), "File Open Error")
            Throw New xmlBuffError
        Finally
            If Not (xreader Is Nothing) Then
                xreader.Close()
            End If
            loadedBuffs.Add(buff_name)
        End Try
        If noError = True Then
            ' put values into system
            old_noRecalc = noRecalc
            noRecalc = True
            For i = 0 To NUM_ATTRS - 1
                attrCA.Item(i).Buff = attrCA.Item(i).Buff + ab(i)
            Next
            For i = 0 To NUM_SKILLS - 1
                skillsCA.Item(i).Buff = skillsCA.Item(i).Buff + sk(i)
            Next
            hpCA.Buff = hpCA.Buff + hp
            npCA.Buff = npCA.Buff + np
            noRecalc = old_noRecalc
            recalcIP()
        End If
    End Sub

    ' load characters buffs from file to string array
    Public Function loadDataDisplay_Buff(ByVal filename As String) As String()
        Dim i As Integer
        Dim xreader As XmlTextReader
        Dim noError As Boolean = True
        Dim buff_name As String
        Dim buff_display As String = ""
        Dim buff_val As Integer

        Try
            buff_name = getNamePart(filename)
            'Create a new XmlTextReader.
            xreader = New XmlTextReader(filename)
            xreader.ReadStartElement("BUFF")
            Do While xreader.IsStartElement()
                If xreader.IsStartElement("ATTR") Then
                    i = CInt(xreader.GetAttribute("ID"))
                    xreader.ReadStartElement("ATTR")
                    buff_val = CInt(xreader.ReadElementString("BUFF"))
                    buff_display &= attrNames(i) & " : +" & CStr(buff_val) & "#"
                    xreader.ReadEndElement()
                End If
                If xreader.IsStartElement("SKILL") Then
                    i = CInt(xreader.GetAttribute("ID"))
                    xreader.ReadStartElement("SKILL")
                    buff_val = CInt(xreader.ReadElementString("BUFF"))
                    buff_display &= skillNames(i) & " : +" & CStr(buff_val) & "#"
                    xreader.ReadEndElement()
                End If
                If xreader.IsStartElement("HP") Then
                    xreader.ReadStartElement("HP")
                    buff_val = CInt(xreader.ReadElementString("BUFF"))
                    buff_display &= "Max Health : +" & CStr(buff_val) & "#"
                    xreader.ReadEndElement()
                End If
                If xreader.IsStartElement("NANO") Then
                    xreader.ReadStartElement("NANO")
                    buff_val = CInt(xreader.ReadElementString("BUFF"))
                    buff_display &= "Max Nano : +" & CStr(buff_val) & "#"
                    xreader.ReadEndElement()
                End If
            Loop
            xreader.ReadEndElement()
        Catch
            ' do error thing here
            noError = False
            MsgBox("There was a problem loading the buff file <" & filename & ">", CType(MsgBoxStyle.Exclamation + MsgBoxStyle.OKOnly, Microsoft.VisualBasic.MsgBoxStyle), "File Open Error")
            Throw New xmlBuffError
        Finally
            If Not (xreader Is Nothing) Then
                xreader.Close()
            End If
        End Try
        Return Split(buff_display, "#")
    End Function

    ' unload characters buffs from file
    Public Sub unloadData_Buff(ByVal filename As String)
        Dim i As Integer
        Dim ab(NUM_ATTRS) As Integer
        Dim sk(NUM_SKILLS) As Integer
        Dim hp As Integer = 0
        Dim np As Integer = 0
        Dim xreader As XmlTextReader
        Dim noError As Boolean = True
        Dim buff_name As String

        For i = 0 To NUM_ATTRS - 1
            ab(i) = 0
        Next
        For i = 0 To NUM_SKILLS - 1
            sk(i) = 0
        Next

        Try
            buff_name = getNamePart(filename)
            'Create a new XmlTextReader.
            xreader = New XmlTextReader(filename)
            xreader.ReadStartElement("BUFF")
            Do While xreader.IsStartElement()
                If xreader.IsStartElement("ATTR") Then
                    i = CInt(xreader.GetAttribute("ID"))
                    xreader.ReadStartElement("ATTR")
                    ab(i) = CInt(xreader.ReadElementString("BUFF"))
                    xreader.ReadEndElement()
                End If
                If xreader.IsStartElement("SKILL") Then
                    i = CInt(xreader.GetAttribute("ID"))
                    xreader.ReadStartElement("SKILL")
                    sk(i) = CInt(xreader.ReadElementString("BUFF"))
                    xreader.ReadEndElement()
                End If
                If xreader.IsStartElement("HP") Then
                    xreader.ReadStartElement("HP")
                    hp = CInt(xreader.ReadElementString("BUFF"))
                    xreader.ReadEndElement()
                End If
                If xreader.IsStartElement("NANO") Then
                    xreader.ReadStartElement("NANO")
                    np = CInt(xreader.ReadElementString("BUFF"))
                    xreader.ReadEndElement()
                End If
            Loop
            xreader.ReadEndElement()
        Catch
            ' do error thing here
            noError = False
            MsgBox("There was a problem unloading the buff file <" & filename & ">", CType(MsgBoxStyle.Exclamation + MsgBoxStyle.OKOnly, Microsoft.VisualBasic.MsgBoxStyle), "File Open Error")
            Throw New xmlBuffError
        Finally
            If Not (xreader Is Nothing) Then
                xreader.Close()
            End If
            loadedBuffs.Remove(buff_name)
        End Try
        If noError = True Then
            ' put values into system
            noRecalc = True
            For i = 0 To NUM_ATTRS - 1
                attrCA.Item(i).Buff = attrCA.Item(i).Buff - ab(i)
            Next
            For i = 0 To NUM_SKILLS - 1
                skillsCA.Item(i).Buff = skillsCA.Item(i).Buff - sk(i)
            Next
            hpCA.Buff = hpCA.Buff - hp
            npCA.Buff = npCA.Buff - np
            noRecalc = False
            recalcIP()
        End If
    End Sub

    ' load characters perks from file
    Public Sub loadData_Perk(ByVal filename As String)
        Dim i As Integer
        Dim ab(NUM_ATTRS) As Integer
        Dim sk(NUM_SKILLS) As Integer
        Dim hp As Integer = 0
        Dim np As Integer = 0
        Dim xreader As XmlTextReader
        Dim noError As Boolean = True
        Dim node_name As String
        Dim attr_value As String

        For i = 0 To NUM_ATTRS - 1
            ab(i) = 0
        Next
        For i = 0 To NUM_SKILLS - 1
            sk(i) = 0
        Next

        Try
            'Create a new XmlTextReader.
            xreader = New XmlTextReader(filename)
            Do While xreader.Read()
                If xreader.NodeType = XmlNodeType.Element Then
                    node_name = xreader.Name
                    If xreader.HasAttributes() Then
                        attr_value = xreader.GetAttribute("NAME")
                    End If
                End If
                If xreader.NodeType = XmlNodeType.Text Then
                    If node_name = "PROF" Then
                        If CInt(xreader.Value) <> fmMain.ProfIndex Then
                            Throw New WrongProfException
                        End If
                    End If
                    If node_name = "BREED" Then
                        If CInt(xreader.Value) <> fmMain.BreedIndex Then
                            Throw New WrongBreedException
                        End If
                    End If
                    If node_name = "LEVEL" Then
                        If CInt(xreader.Value) > fmMain.NudLevelValue Then
                            Throw New LevelTooLowException
                        End If
                    End If
                    If node_name = "AILEVEL" Then
                        If CInt(xreader.Value) > fmMain.NudAlienTitleValue Then
                            Throw New AILevelTooLowException
                        End If
                    End If
                    If node_name = "PERK" Then
                        AddPerk(GetPerkIndex(attr_value, CInt(xreader.Value)))
                    End If
                    If node_name = "AIPERK" Then
                        AddAlienPerk(GetAlienPerkIndex(attr_value, CInt(xreader.Value)))
                    End If
                    If node_name = "RESEARCH" Then
                        AddResearch(GetResearchIndex(attr_value, CInt(xreader.Value)))
                    End If
                End If
                If xreader.NodeType = XmlNodeType.EndElement Then
                    If xreader.Name = "PERKLIST" Then
                        Exit Do
                    End If
                End If
            Loop
        Catch wpe As WrongProfException
            ' do error thing here
            noError = False
            MsgBox("This perk file is not intended for the currently selected profession", CType(MsgBoxStyle.Exclamation + MsgBoxStyle.OKOnly, Microsoft.VisualBasic.MsgBoxStyle), "File Open Error")
        Catch wpe As WrongBreedException
            ' do error thing here
            noError = False
            MsgBox("This perk file is not intended for the currently selected breed", CType(MsgBoxStyle.Exclamation + MsgBoxStyle.OKOnly, Microsoft.VisualBasic.MsgBoxStyle), "File Open Error")
        Catch
            ' do error thing here
            noError = False
            MsgBox("There was a problem loading the perk file <" & filename & ">", CType(MsgBoxStyle.Exclamation + MsgBoxStyle.OKOnly, Microsoft.VisualBasic.MsgBoxStyle), "File Open Error")
        Finally
            If Not (xreader Is Nothing) Then
                xreader.Close()
            End If
        End Try
        If noError = True Then
            ' put values into system
            ConfigurePerks()
            recalcIP()
        End If
    End Sub

    ' load characters misc from file
    Public Sub loadData_Misc(ByVal filename As String)
        Dim i As Integer
        Dim id As Integer
        Dim ql As Integer
        Dim xreader As XmlTextReader
        Dim noError As Boolean = True
        Dim ba2_id() As Boolean
        Dim ba2_ql() As Integer

        ReDim ba2_id(num_Benefits2)
        ReDim ba2_ql(num_Benefits2)
        For i = 0 To num_Benefits2
            ba2_id(i) = False
            ba2_ql(i) = 0
        Next
        Try
            'Create a new XmlTextReader.
            xreader = New XmlTextReader(filename)
            xreader.ReadStartElement("ORGBENEFIT")
            If xreader.EOF = False Then
                Do
                    If xreader.IsStartElement("ITEM") Then
                        id = CInt(xreader.GetAttribute("ID"))
                        ql = CInt(xreader.GetAttribute("QL"))
                        ba2_id(id) = True
                        ba2_ql(id) = ql
                        xreader.Read()
                    Else
                        Exit Do
                    End If
                Loop
                xreader.ReadEndElement()
            End If
        Catch
            ' do error thing here
            noError = False
            MsgBox("There was a problem loading the org benefit file <" & filename & ">", CType(MsgBoxStyle.Exclamation + MsgBoxStyle.OKOnly, Microsoft.VisualBasic.MsgBoxStyle), "File Open Error")
        Finally
            If Not (xreader Is Nothing) Then
                xreader.Close()
            End If
        End Try
        If noError = True Then
            ' put values into system
            ClearMisc()
            For i = 0 To num_Benefits2
                BenefitArray2(i).Selected = ba2_id(i)
                BenefitArray2(i).SelectedQL = ba2_ql(i)
            Next
            processAllBenefits()
            recalcIP()
        End If
    End Sub

    Function getNamePart(ByVal in_name As String) As String
        Return System.IO.Path.GetFileNameWithoutExtension(in_name)
    End Function

End Module
